package build

import (
	"fmt"
	"strconv"

	"git.akyoto.dev/cli/q/src/build/arch/x64"
	"git.akyoto.dev/cli/q/src/build/asm"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/fs"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/go/color/ansi"
)

// Function represents a function.
type Function struct {
	Name      string
	File      *fs.File
	Head      token.List
	Body      token.List
	Variables map[string]*Variable
	Assembler asm.Assembler
	Error     error
}

// Compile turns a function into machine code.
func (f *Function) Compile() {
	f.Assembler.Label(f.Name)
	start := 0
	groupLevel := 0

	for i, t := range f.Body {
		if start == i && (t.Kind == token.NewLine || t.Kind == token.BlockStart || t.Kind == token.BlockEnd) {
			start = i + 1
			continue
		}

		switch t.Kind {
		case token.NewLine:
			if groupLevel > 0 {
				continue
			}

			if start != -1 {
				instruction := f.Body[start:i]
				err := f.CompileInstruction(instruction)

				if err != nil {
					f.Error = err
					return
				}

				start = -1
			}

			start = i + 1

		case token.GroupStart:
			groupLevel++

		case token.GroupEnd:
			groupLevel--

		case token.BlockStart:
			// Add scope

		case token.BlockEnd:
			// Remove scope
		}
	}

	f.Assembler.Return()
}

// PrintAsm shows the assembly instructions.
func (f *Function) PrintAsm() {
	ansi.Bold.Println(f.Name)
	ansi.Dim.Println("╭────────────────────────────────────────────────────────────")

	for _, x := range f.Assembler.Instructions {
		ansi.Dim.Print("│ ")
		fmt.Print(x.Mnemonic.String())

		if x.Data != nil {
			fmt.Print(" " + x.Data.String())
		}

		fmt.Print("\n")
	}

	ansi.Dim.Println("╰────────────────────────────────────────────────────────────")
}

// CompileInstruction compiles a single instruction.
func (f *Function) CompileInstruction(line token.List) error {
	if len(line) == 0 {
		return nil
	}

	if line[0].Kind == token.Keyword {
		switch line[0].Text() {
		case "return":
			if len(line) > 1 {
				value := expression.Parse(line[1:])
				defer value.Close()
				// TODO: Set the return value
			}

			f.Assembler.Return()

		default:
			return errors.New(&errors.KeywordNotImplemented{Keyword: line[0].Text()}, f.File, line[0].Position)
		}
	}

	expr := expression.Parse(line)

	if expr == nil {
		return nil
	}

	defer expr.Close()

	if expr.Token.Kind == token.Number || expr.Token.Kind == token.Identifier || expr.Token.Kind == token.String {
		return errors.New(&errors.InvalidInstruction{Instruction: expr.Token.Text()}, f.File, expr.Token.Position)
	}

	if expr.Token.Text() == ":=" {
		if len(expr.Children) < 2 {
			return errors.New(errors.MissingAssignValue, f.File, expr.LastChild().Token.After())
		}

		name := expr.Children[0]
		value := expr.Children[1]

		expr.RemoveChild(value)

		f.Variables[name.Token.Text()] = &Variable{
			Name:    name.Token.Text(),
			Value:   value,
			IsConst: true,
		}

		return nil
	}

	if expr.Token.Text() == "λ" {
		funcName := expr.Children[0].Token.Text()
		parameters := expr.Children[1:]

		for i, parameter := range parameters {
			switch parameter.Token.Kind {
			case token.Identifier:
				name := parameter.Token.Text()
				variable, exists := f.Variables[name]

				if !exists {
					panic("Unknown identifier " + name)
				}

				if !variable.IsConst {
					panic("Not implemented yet")
				}

				n, _ := strconv.Atoi(variable.Value.Token.Text())
				f.Assembler.MoveRegisterNumber(x64.SyscallArgs[i], uint64(n))

			case token.Number:
				value := parameter.Token.Text()
				n, _ := strconv.Atoi(value)
				f.Assembler.MoveRegisterNumber(x64.SyscallArgs[i], uint64(n))

			default:
				panic("Unknown expression")
			}
		}

		if funcName == "syscall" {
			f.Assembler.Syscall()
		} else {
			f.Assembler.Call(funcName)
		}

		return nil
	}

	return nil
}

// String returns the function name.
func (f *Function) String() string {
	return f.Name
}
