package asm

import (
	"encoding/binary"

	"git.akyoto.dev/cli/q/src/build/arch/x64"
)

// Assembler contains a list of instructions.
type Assembler struct {
	Instructions []Instruction
}

// New creates a new assembler.
func New() *Assembler {
	return &Assembler{
		Instructions: make([]Instruction, 0, 8),
	}
}

// Finalize generates the final machine code.
func (a *Assembler) Finalize() ([]byte, []byte) {
	code := make([]byte, 0, len(a.Instructions)*8)
	data := make([]byte, 0, 16)
	labels := map[string]Address{}
	pointers := []Pointer{}

	for _, x := range a.Instructions {
		switch x.Mnemonic {
		case MOVE:
			regNum := x.Data.(*RegisterNumber)
			code = x64.MoveRegNum32(code, uint8(regNum.Register), uint32(regNum.Number))

		case RETURN:
			code = x64.Return(code)

		case SYSCALL:
			code = x64.Syscall(code)

		case CALL:
			code = x64.Call(code, 0x00_00_00_00)
			label := x.Data.(*Label)
			nextInstructionAddress := len(code)

			pointers = append(pointers, Pointer{
				Position: Address(len(code) - 4),
				Resolve: func() Address {
					destination := labels[label.Name]
					distance := int32(destination) - int32(nextInstructionAddress)
					return Address(distance)
				},
			})

		case LABEL:
			labels[x.Data.(*Label).Name] = Address(len(code))

		default:
			panic("Unknown mnemonic: " + x.Mnemonic.String())
		}
	}

	// dataStart := config.BaseAddress + config.CodeOffset + Address(len(code))

	for _, pointer := range pointers {
		slice := code[pointer.Position : pointer.Position+4]
		binary.LittleEndian.PutUint32(slice, pointer.Resolve())
	}

	return code, data
}

// Merge combines the contents of this assembler with another one.
func (a *Assembler) Merge(b *Assembler) {
	a.Instructions = append(a.Instructions, b.Instructions...)
}
